from main import UNSET_COORD
from utils.datatypes import *
from utils.Struct import Struct
from utils import units
from DataTarget import DataTarget

from urllib import unquote # for filename conversion

import gtk


#
# Abstract class for DisplayTargets.
#
class DisplayTarget(DataTarget):

    # observer actions
    OBS_GEOMETRY = 0

    # user actions
    ACTION_CLICK = "click"
    ACTION_DOUBLECLICK = "doubleclick"
    ACTION_PRESS = "press"
    ACTION_RELEASE = "release"
    ACTION_MENU = "menu"
    ACTION_SCROLL = "scroll"
    ACTION_ENTER = "enter"
    ACTION_LEAVE = "leave"
    ACTION_MOTION = "motion"
    ACTION_FILE_DROP = "file-drop"
    ACTION_KEY_PRESS = "key-press"
    ACTION_KEY_RELEASE = "key-release"

    # placement anchors
    ANCHOR_NW = "nw"
    ANCHOR_N = "n"
    ANCHOR_NE = "ne"
    ANCHOR_E = "e"
    ANCHOR_SE = "se"
    ANCHOR_S = "s"
    ANCHOR_SW = "sw"
    ANCHOR_W = "w"
    ANCHOR_CENTER = "center"


    # what we accept for drag & drop
    __DND_FILE = [('text/uri-list', 0, 0)]


    def __init__(self, parent):

        # the real coordinates and size of this widget
        self.__geometry = (0, 0, 0, 0)
        self.__old_geometry = (UNSET_COORD, UNSET_COORD, 0, 0)

        # coordinates and size of this widget as set by the user
        self.__user_geometry = (UNSET_COORD, UNSET_COORD,
                                UNSET_COORD, UNSET_COORD)

        # the units of the geometry values
        self.__geometry_units = (units.UNIT_PX, units.UNIT_PX,
                                 units.UNIT_PX, units.UNIT_PX)

        # the widget to which this one is positioned relatively
        self.__relative = None

        # relative settings (is_rel_x, is_rel_y)
        self.__is_relative = (0, 0)

        # the placement anchor
        self.__anchor = self.ANCHOR_NW

        # the value of the last notify_handle_action call
        self.__had_action = 0

        DataTarget.__init__(self, parent)

        for prop, datatype in [("x", TYPE_COORD_SIZE),
                               ("y", TYPE_COORD_SIZE),
                               ("width", TYPE_COORD_SIZE),
                               ("height", TYPE_COORD_SIZE)]:
            self._register_property(prop, datatype,
                                    self._setp_geometry, self._getp_geometry)
        #end for

        self._register_property("relative-to", TYPE_LIST,
                                self._setp_relative_to, self._getp)

        self._register_property("anchor", TYPE_STRING,
                                self._setp_anchor, self._getp)

        self._register_property("visible", TYPE_BOOL,
                                self._setp_visible, self._getp)
        self._setp("visible", True)

        for action in (self.ACTION_ENTER,
                       self.ACTION_LEAVE,
                       self.ACTION_CLICK,
                       self.ACTION_PRESS,
                       self.ACTION_RELEASE,
                       self.ACTION_SCROLL,
                       self.ACTION_FILE_DROP,
                       self.ACTION_DOUBLECLICK,
                       self.ACTION_MOTION,
                       self.ACTION_MENU,
                       self.ACTION_KEY_PRESS,
                       self.ACTION_KEY_RELEASE):
            self._register_action(action)
        #end for

        #gtk.idle_add(gtk.Widget.show, self.get_widget())
        #self.get_widget().set_double_buffered(gtk.FALSE)


    #
    # Reacts on notifications by the relative of this widget.
    #
    def set_relative_to(self, relative):

        self.__relative = relative
        self.adjust_geometry()



    def __on_file_drop( self, widget, context, x, y, data, info, time ):
        ''' catch DND events and process them to send them to them
        main display, which forwards them to the sensor '''

        # get the display
        display = self._get_display()

        # tell the main display to send files and coordinates to the sensor
        files = [unquote(uri) for uri in data.data.split("\r\n") if uri != '']
        display.send_action(self, self.ACTION_FILE_DROP,
                            Struct(files = files, _args = [files]))


    #
    # Returns whether this target is standalone, i.e. needs no parent.
    #
    def is_standalone(self): return False


    #
    # Returns the widget of this target.
    #
    def get_widget(self): raise NotImplementedError


    #
    # Returns the true coordinates of this target when the given coordinates
    # are the hotspot.
    #
    def get_anchored_coords(self, x, y, w, h):

        if (x == y == UNSET_COORD): return (x, y)

        anchor = self.__anchor
        if (anchor in (self.ANCHOR_NW, self.ANCHOR_W, self.ANCHOR_SW)):
            ax = x
        elif (anchor in (self.ANCHOR_N, self.ANCHOR_CENTER, self.ANCHOR_S)):
            ax = x - (w / 2)
        else:
            ax = x - w

        if (anchor in (self.ANCHOR_NW, self.ANCHOR_N, self.ANCHOR_NE)):
            ay = y
        elif (anchor in (self.ANCHOR_W, self.ANCHOR_CENTER, self.ANCHOR_E)):
            ay = y - (h / 2)
        else:
            ay = y - h

        return (ax, ay)


    #
    # Returns the position of the anchor in this target. Origin is the top
    # left corner.
    #
    def get_anchor(self):

        x, y, w, h = self.__geometry
        ax, ay = self.get_anchored_coords(x, y, w, h)
        dx = (x - ax)
        dy = (y - ay)

        return (dx, dy)



    #
    # Returns the target ID to which this target is placed relatively.
    #
    def get_relative_to(self):

        return self.__relative



    #
    # Returns the geometry (coordinates and size) of this target.
    #
    def get_geometry(self):

        x, y, w, h = self.__geometry
        if (self.get_prop("visible")):
            return (x, y, w, h)
        else:
            return (x, y, 0, 0)



    #
    # Sets the user geometry of this target.
    #
    def set_user_geometry(self, x = None, y = None,
                          width = None, height = None):

        ox, oy, ow, oh = self.__user_geometry
        if (x == None): x = ox
        if (y == None): y = oy
        if (width == None): width = ow
        if (height == None): height = oh
        self.__user_geometry = (x, y, width, height)



    #
    # Returns the geometry from the user's point of view.
    #
    def get_user_geometry(self):

        return self.__user_geometry


    #
    # Sets the unit of geometry values.
    #
    def __set_geometry_unit(self, x = None, y = None,
                            width = None, height = None):

        ux, uy, uw, uh = self.__geometry_units
        if (x == None): x = ux
        if (y == None): y = uy
        if (width == None): width = uw
        if (height == None): height = uh
        self.__geometry_units = (x, y, width, height)


    #
    # Returns the geometry units.
    #
    def get_geometry_units(self):

        return self.__geometry_units


    #
    # Computes all unit values.
    #
    def _compute_units(self, x = None, y = None, width = None, height = None):

        parent = self._get_parent()
        if (parent != self._get_display()):
            px, py, pw, ph = parent.get_container_geometry()
        else:
            px, py, pw, ph = \
                (0, 0, gtk.gdk.screen_width(), gtk.gdk.screen_height())

        pw = max(1, pw)
        ph = max(1, ph)

        if (x != None):
            pixels = x
            percents = (pixels  / float(pw)) * 100.0
        elif (y != None):
            pixels = y
            percents = (pixels  / float(ph)) * 100.0
        elif (width != None):
            pixels = width
            percents = (pixels  / float(pw)) * 100.0
        elif (height != None):
            pixels = height
            percents = (pixels  / float(ph)) * 100.0
            
        cms, inches, pts = units.pixels_to_all(pixels)

        return {units.UNIT_PX: pixels,
                units.UNIT_PERCENT: percents,
                units.UNIT_CM: cms,
                units.UNIT_IN: inches,
                units.UNIT_PT: pts}


    #
    # Adjusts the geometry of this target.
    #
    def adjust_geometry(self):

        x, y, w, h = self.__geometry
        ux, uy, uw, uh = self.get_user_geometry()
        unit_x, unit_y, unit_w, unit_h = self.get_geometry_units()
        anchor_x, anchor_y = self.get_anchor()
        parent = self._get_parent()

        if (parent != self._get_display()):
            parent_width, parent_height = \
                 self._get_parent().get_container_geometry()[2:]
        else:
            parent_width = gtk.gdk.screen_width()
            parent_height = gtk.gdk.screen_height()

        if (ux == UNSET_COORD): new_x = x + anchor_x
        elif (unit_x == units.UNIT_PERCENT):
            new_x = int(ux * (parent_width / 100.0))
        else: new_x = units.units_to_pixels(ux, unit_x)

        if (uy == UNSET_COORD): new_y = y + anchor_y
        elif (unit_y == units.UNIT_PERCENT):
            new_y = int(uy * (parent_height / 100.0))
        else: new_y = units.units_to_pixels(uy, unit_y)

        if (uw == UNSET_COORD): new_w = self._get_value_for_unset(width = w)
        elif (unit_w == units.UNIT_PERCENT):
            new_w = int(uw * (parent_width / 100.0))
        else: new_w = units.units_to_pixels(uw, unit_w)

        if (uh == UNSET_COORD): new_h = self._get_value_for_unset(height = h)
        elif (unit_h == units.UNIT_PERCENT):
            new_h = int(uh * (parent_height / 100.0))
        else: new_h = units.units_to_pixels(uh, unit_h)

        if (self.__relative):
            is_rel_x, is_rel_y = self.__is_relative

            # get the coords of the anchor of the relative
            anchor_x, anchor_y = self.__relative.get_anchor()
            rx, ry, rw, rh = self.__relative.get_geometry()

            # get the remaining amount of pixels from the anchor to the
            # bottom/right edge of the relative
            tw = rw - anchor_x
            th = rh - anchor_y

            if (ux != UNSET_COORD): new_x += rx + anchor_x # + ux
            else: new_x = rx + anchor_x
            if (uy != UNSET_COORD): new_y += ry + anchor_y #+ uy
            else: new_y = ry + anchor_y
            if (is_rel_x): new_x += tw
            if (is_rel_y): new_y += th
        #end if


        # get coords of the top left corner
        new_x, new_y = self.get_anchored_coords(new_x, new_y, new_w, new_h)
        if ((parent == self._get_display() and self.get_widget().window)
              or (new_x, new_y, new_w, new_h) != self.__old_geometry):
            self.__geometry = (new_x, new_y, new_w, new_h)
            self.__old_geometry = self.__geometry
            self.get_widget().set_size_request(new_w, new_h)
            # if w and h of a widget are 0, we might not propagate the GEOMETRY
            # the problem is, if they are invisible, they have to be notified,
            # otherwise other widgets are taking their place
            self._propagate_geometry()



    #
    # Returns the geometry value that should be used for unset values.
    #
    def _get_value_for_unset(self, x = None, y = None,
                             width = None, height = None):

        if (x != None): return x
        elif (y != None): return y
        elif (width != None): return width
        elif (height != None): return height



    #
    # Tells other targets about geometry changes.
    #
    def _propagate_geometry(self):

        parent = self._get_parent()
        if (parent): gtk.timeout_add(0, parent.adjust_geometry)
        self.update_observer(self.OBS_GEOMETRY)


    #
    # Sets the position of this target.
    #
    def set_position(self, x, y):

        ox, oy, w, h = self.__geometry
        if ((x, y) != (ox, oy)):
            self.__geometry = (x, y, w, h)
            self.adjust_geometry()



    #
    # Sets the size of this target. Use this instead of set_size_request() in
    # targets to set the size manually.
    #
    def set_size(self, width, height):

        x, y, w, h = self.__geometry
        if ((w, h) != (width, height)):
            self.__geometry = (x, y, width, height)
            self.adjust_geometry()



    def handle_action(self, action, px, py, event):

        if (action in [self.ACTION_MOTION]):
            x, y = self.get_widget().get_pointer()
            event = Struct(x = x, y = y, _args = [x, y])

        DataTarget.handle_action(self, action, px, py, event)



    #
    # Notifies the target whether the action occured on it or not.
    # This detects ENTER and LEAVE events.
    #
    def notify_handle_action(self, value):

        # enter notify
        if (not self.__had_action and value):
            action = self.ACTION_ENTER
            if (self.get_action_call(action)):
                self.handle_action(action, 0, 0, Struct())

        # leave notify
        elif (self.__had_action and not value):
            action = self.ACTION_LEAVE
            if (self.get_action_call(action)):
                self.handle_action(action, 0, 0, Struct())

        self.__had_action = value



    #
    # Geometry properties.
    #
    def _setp_geometry(self, key, value):

        if (key == "x"):
            v, unit = value
            self.set_user_geometry(x = v)
            self.__set_geometry_unit(x = unit)
            self.adjust_geometry()
            self._setp(key, value)

        elif (key == "y"):
            v, unit = value
            self.set_user_geometry(y = v)
            self.__set_geometry_unit(y = unit)
            self.adjust_geometry()
            self._setp(key, value)

        elif (key == "width"):
            v, unit = value
            self.set_user_geometry(width = v)
            self.__set_geometry_unit(width = unit)
            self.adjust_geometry()
            self._setp(key, value)

        elif (key == "height"):
            v, unit = value
            self.set_user_geometry(height = v)
            self.__set_geometry_unit(height = unit)
            self.adjust_geometry()
            self._setp(key, value)


    def _setp_relative_to(self, key, value):

        name, mode = value
        if (mode == "x"): self.__is_relative = (1, 0)
        elif (mode == "y"): self.__is_relative = (0, 1)
        elif (mode == "xy"): self.__is_relative = (1, 1)
        
        self._get_parent().set_relative(self, name)
        self._setp(key, value)


    def _setp_anchor(self, key, value):

        x, y = self.get_user_geometry()[:2]
        self.__anchor = value
        self.adjust_geometry()
        self._setp(key, value)


    def _getp_geometry(self, key):

        x, y, w, h = self.get_geometry()
        x, y = self.get_anchored_coords(x, y, w, h)
        if (key == "x"): return self._compute_units(x = x)
        elif (key == "y"): return self._compute_units(y = y)
        elif (key == "width"): return self._compute_units(width = w)
        elif (key == "height"): return self._compute_units(height = h)

        
    #
    # "visible" property.
    #
    def _setp_visible(self, key, value):

        if (value): self.get_widget().show()
        else: self.get_widget().hide()
        self._setp(key, value)
        self.adjust_geometry()
        self._propagate_geometry()



    #
    # Action handlers.
    #
    def _setp__action(self, key, value):

        DataTarget._setp__action(self, key, value)
        if (key == "on-file-drop"):
            self.get_widget().drag_dest_set(gtk.DEST_DEFAULT_ALL,
                                            self.__DND_FILE,
                                            gtk.gdk.ACTION_COPY)
            self.get_widget().connect("drag_data_received",
                                      self.__on_file_drop)
