#
# Base class for controls. Every control has to be derived from this class.
#
class Control(object):

    def __init__(self):

        # the handlers for property bindings
        self.__handlers = {}

        # flag indicating whether the control has been stopped
        self.__is_stopped = False


    def _update(self, property):
        """
        Issues a change of the given property.
        """

        lst = self.__handlers.get(property, [])
        for handler, args in lst:
            handler(getattr(self, property), *args)



    def bind(self, property, handler, *args):
        """
        Binds a handler to property changes.
        """

        self.__handlers.setdefault(property, []).append((handler, args))



    def _add_timer(self, interval, callback, *args):
        """
        Runs a timer callback at the given intervals.
        """

        def f():
            ret = callback(*args)
            if (ret and not self.__is_stopped): return True
            else: return False
            
        import gtk
        gtk.timeout_add(interval, f)

        

    #
    # Stops the control.
    #
    def stop(self):

        self.__is_stopped = True
        # remove bound handlers
        self.__handlers.clear()
        self._shutdown()


    #
    # Shutdown handler for controls. Controls should override this method,
    # if needed.
    #
    def _shutdown(self): pass


    def __interface(self):
        """
        Returns a description of the interface of this control.
        """

        out = "Interface:\n"
        out += "-" * 80 + "\n"
        items = self.__class__.__dict__.items()
        items.sort()
        for key, value in items:
            if (type(value) == property):
                if (value.fget): r = "r"
                else: r = ""
                length = len(key)
                out += " " + key + " " * (25 - length)
                if (value.fget): out += "r "
                else: out += "  "
                if (value.fset): out += "w"
                else: out += " "
                out += "   " + (value.__doc__ or "no description") + "\n"
        return out

    interface = property(__interface, doc = "Interface description")


