import Permission

import types


#
# Abstract base class for interfaces.
#
class Interface(object):

    def __get_members(impl):

        return [(name, type(member), member) for name, member
                in impl.__dict__.items()]

    __get_members = classmethod(__get_members)
    

    #
    # Returns a dictionary of the properties of the given implementation.
    #
    def get_properties(iface):

        return dict([(n, p) for n, t, p in iface.__get_members()
                      if (t == property)])

    get_properties = classmethod(get_properties)



    #
    # Returns a list of the methods of the given implementation.
    #
    def get_methods(iface):

        return [(n, "") for n, t, m in iface.__get_members()
                if (t == types.FunctionType)]

    get_methods = classmethod(get_methods)



    #
    # Returns a list of the permissions of the properties.
    #
    def get_permissions(iface):

        props = [(k, v) for k, v in iface.get_properties()
                 if (type(k) == property and k not in Interface.__dict__)]

        for nil, p  in props:
            assert isinstance(p, Permission._Permission)

        return props

    get_permissions = classmethod(get_permissions)




    #
    # Static method for returning the unique identifier of the given interface.
    #
    def get_id(iface):

        # the algorithm is simple:
        # build a string containing the class name and a sorted list of the
        # properties and return the MD5 fingerprint of it
        import md5
        import utils
        
        name = iface.__name__
        items1 = iface.get_permissions()
        items1.sort() # we need python2.4
        items2 = iface.get_methods()
        items2.sort()
        tmp = ["%s:%s" % (k, v) for k, v in items1 + items2]
    
        tmp = "%s:%s" % (name, ",".join(tmp))
        print tmp

        ash = md5.new(tmp).hexdigest()
        # encode it in base36 to get shorter ID
        ash = utils.radix( int(ash, 16), 36 )
        # ID have to be 25 char long
        ash = '0' * (25 - len(ash)) + ash

        assert len(ash) == 25
        
        return "%s:%s" % (name, ash)
    
    get_id = classmethod(get_id)



    #
    # Returns the interfaces implemented by the given class.
    #
    def get_interfaces(impl):

        return [c for c in impl.__bases__
                if issubclass(c, Interface) and c != Interface ]
    
    get_interfaces = staticmethod(get_interfaces)

    

    #
    # Static method for asserting that the interfaces are implemented properly
    # in the given class. Throws exceptions otherwise.
    #
    def assert_interfaces(impl):
        
        # find the implemented properties
        impl_props = dict(impl.get_properties())

        # find the implemented methods
        impl_meths = dict(impl.get_methods())

        # check all interfaces
        for iface in Interface.get_interfaces(impl):

            # check properties
            required = iface.get_permissions()
            for name, perm in required:
                if name not in impl_props:
                    raise NotImplementedError("Missing implementation for "
                                              "%s.%s" % (iface.__name__, name))
       
                # check if the property's mode is according to the interfaces
                if (perm != Permission.Permission(implemented[name])):
                    raise NotImplementedError("Broken implementation for "
                                      "%s.%s" % (iface.__name__, name))
            #end for

            # check methods
            required = iface.get_methods()
            for name, meth in required:    
                if name not in impl_meths:
                    raise NotImplementedError("Missing implementation for "
                                              "%s.%s" % (iface.__name__, name))
            #end for
            
        #end for
        
    assert_interfaces = staticmethod(assert_interfaces)



if __name__ == '__main__':


    import sys; sys.path.append('.')
    
    import plugin.Permission as Permission


    class A(Interface):
        a = Permission.READ

    class B(Interface):
        b = Permission.READWRITE
        c = Permission.WRITE


    class Foo(A, B):

        a = property(int)
        b = property(int, int)
        c = property(fset = int)

        def test(self): pass
        

    print A.get_permissions()
    print A.get_id()
    print
    print B.get_permissions()
    print B.get_id()
    print
    print Foo.get_properties()
    print "methods", Foo.get_methods()
    print Interface.get_interfaces(Foo)
    Interface.assert_interfaces(Foo)
